from typing import List


class Tokenizer:
    @staticmethod
    def get_words(text: str) -> List[str]:
        """finds words.
        not included: numbers.
        included: hyphenated words, contractions
        """
        text = text.strip()
        text = text.lower()
        # we're lowering now, so we don't need the caps
        alpha = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        joiners = "'-"
        words = []
        word_start = 0
        pointer = -1

        def back(index):
            while index > -1 and text[index - 1] not in alpha:
                index = index - 1
            return index

        def forward(index):
            while index < len(text) - 1 and text[index] not in alpha:
                index = index + 1
            return index

        def take_a_word(f, b, call):
            # shouldn't happen. but if it does, -1 wraps.
            if b == -1:
                # print("b is -1. returning ''")
                return ""
            fw = forward(f)
            bk = back(b)
            # can happen on back-to-back special chars. if it does, -1 wraps.
            if bk == -1:
                # print("bk is -1. returning ''")
                return ""

            word = text[fw:bk]
            word = word.strip()
            if word != "":
                words.append(word)
            """
            print(f"\ncall: {call}")
            print(  f"f,b: {f},{b}")
            print(  f"fw,bk: {fw},{bk}")
            print(  f"found word: {word} ")
            print(  f"in {text}")
            """
            return word

        for c in text:
            pointer = pointer + 1
            if c == " " and pointer > word_start:
                take_a_word(word_start, pointer, 41)
                word_start = pointer + 1
            elif c in alpha:
                continue
            elif c in joiners:
                if pointer == 0:
                    continue
                elif pointer == len(text) - 1:
                    continue
                elif text[pointer - 1] in alpha and text[pointer + 1] in alpha:
                    continue
                else:
                    if pointer - 1 > word_start:
                        take_a_word(word_start, pointer, 54)
                        word_start = pointer
                    else:
                        word_start = pointer
                        continue
            else:
                continue

        if word_start < pointer:
            take_a_word(word_start, pointer + 1, 63)

        return words



from dataclasses import dataclass, field
from typing import List, Optional, Dict
from . import SectionMetadata
from . import Files
from . import Metadata


@dataclass
class BookMetadata(Metadata):
    FILES: Files
    SECTIONS: List[SectionMetadata] = field(default_factory=list)
    TITLE: str = ""
    AUTHOR: str = ""

    def __post_init__(self):
        self.seek_metadata()

    def seek_metadata(self) -> None:
        if self.FILES is None or self.FILES.INPUT is None:
            return
        if self.TITLE == "" or self.TITLE is None:
            title = self._seek("TITLE")
            if title is not None:
                self.TITLE = title
        if self.AUTHOR == "" or self.AUTHOR is None:
            author = self._seek("AUTHOR")
            if author is not None:
                self.AUTHOR = author

    def _seek(self, token: str) -> Optional[str]:
        if token not in ["TITLE", "AUTHOR"]:
            raise Exception("token must be either title or author")
        with open(self.FILES.INPUT) as file:
            for line in file:
                if line[0] == "#" and line.find(token) > 0:
                    return line[(line.find(token) + len(token) + 1) :].strip()
        return None

    @property
    def count(self) -> int:
        return len(self.SECTIONS)

    @property
    def stand_alone_count(self) -> int:
        count = 0
        for s in self.SECTIONS:
            if s.STAND_ALONE:
                count = count + 1
        return count

    def new_section(self):
        section = SectionMetadata()
        self.SECTIONS.append(section)
        return section

    @property
    def word_count(self) -> int:
        words = 0
        for s in self.SECTIONS:
            words += s.WORD_COUNT
        return words

    def unique_words_count(self) -> int:
        return len(self.words())

    def words(self) -> Dict[str, int]:
        words: Dict[str, int] = {}
        for s in self.SECTIONS:
            for word in s.WORDS:
                if word in words:
                    words[word] = words[word] + 1
                else:
                    words[word] = 1
        return words



from dataclasses import dataclass, field
from typing import List, Optional
from . import Markup
from . import Metadata


class SectionContentException(Exception):
    pass


@dataclass
class SectionMetadata(Metadata):
    FILE: Optional[str] = None
    RELATIVE_PATH: Optional[str] = None
    NAME: Optional[str] = None
    FIRST_LINE: Optional[str] = None
    FILE_LINE_COUNT: int = 0
    STAND_ALONE: bool = False
    CHAR_COUNT: int = 0
    WORD_COUNT: int = 0
    WORDS: List[str] = field(default_factory=list)

    def unique_words_count(self) -> int:
        words = {}
        for w in self.WORDS:
            words[w] = w
        return len(words)

    def set_first_line(self, markup: Markup, line: str) -> None:
        if markup is None:
            raise SectionContentException("Markup cannot be None")
        if line is None:
            raise SectionContentException("Line cannot be None")
        line = line.strip()
        if (
            line.find(markup.NEW_SECTION) == 0
            or line.find(markup.CHAPTER_TITLE) == 0
            or line.find(markup.BOOK_TITLE) == 0
        ):
            self.STAND_ALONE = True
        self.FIRST_LINE = line

    def add_relative_file(self, name: str) -> None:
        if name is None:
            raise SectionContentException("Name cannot be None")
        self.RELATIVE_PATH = name
        slash = name.rfind("/")
        if slash > -1:
            name = name[slash + 1 :]
        dot = name.rfind(".")
        if dot > -1:
            name = name[0:dot]
        self.NAME = name



from configparser import RawConfigParser
from dataclasses import dataclass
from os import path


@dataclass
class Files:
    INPUT: str = "contents.csv"
    OUTPUT: str = "./out/out.docx"
    FILES: str = "../texts"
    CONFIG: str = "config/files.ini"

    def __post_init__(self):
        self._config = RawConfigParser()
        self._load_config()

    def reload(self):
        self._load_config()

    def _load_config(self):
        if path.isfile(self.CONFIG):
            self._config.read(self.CONFIG)
            section = "files"
            try:
                self.INPUT = self._config[section]["input"]
            except KeyError:
                pass
            try:
                self.OUTPUT = self._config[section]["output"]
            except KeyError:
                pass
            try:
                self.FILES = self._config[section]["files"]
            except KeyError:
                pass
        else:
            print(f"No files {self.CONFIG} file found. Using default file locations.")



import csv
from . import Files
from . import Markup
from . import Fonts
from . import Section
from . import RegularSection
from . import MetadataSection
from . import BookMetadata
from docx import Document
import traceback
from typing import List, Optional
import os
from docxcompose.composer import Composer


class UnknownBookfileTypeException(Exception):
    pass


class BookfileNotFoundException(Exception):
    pass


class BookfileMistakeException(Exception):
    pass


class Book:
    def __init__(
        self,
        files: Files,
        markup: Optional[Markup],
        fonts: Optional[Fonts],
        document: Document,
    ) -> None:
        self._files = files
        self._fonts = fonts
        self._markup = markup
        self._file_count = 0
        self._document = document
        self._metadata = BookMetadata(FILES=files)

    @property
    def files(self) -> int:
        return self._file_count

    @property
    def metadata(self) -> BookMetadata:
        return self._metadata

    def create(self) -> int:
        """returns the rows count, including comments but not blanks"""
        cnt = 0
        rows = self._get_rows()
        for row in rows:
            if len(row[0]) > 0 and row[0][0] == "#":
                self._check_for_inserts(row)
                self._check_for_metadata(row)
            else:
                try:
                    # build the book here
                    self._append_section(row)
                    self._file_count = self._file_count + 1
                except Exception as e:
                    print(f"Error: {cnt}: {row}: {e}")
                    print(traceback.format_exc())
            cnt = cnt + 1
        return cnt

    def _get_rows(self) -> List[str]:
        try:
            if self._files.INPUT.endswith(".csv"):
                return self._get_rows_csv()
            elif self._files.INPUT.endswith(".txt") or self._files.INPUT.endswith(
                ".bookfile"
            ):
                return self._get_rows_txt()
            else:
                raise UnknownBookfileTypeException(
                    f"Unknown bookfile type. Please check {self._files.INPUT}"
                )
        except FileNotFoundError:
            raise BookfileNotFoundException(
                f"Bookfile not found. Please check {self._files.INPUT}"
            )

    def _get_rows_csv(self) -> List[str]:
        rows = []
        with open(self._files.INPUT) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            for row in csv_reader:
                if len(row) == 0:
                    pass
                elif len(row[0]) > 0 and row[0][0] == "#":
                    rows.append(row[0])
                else:
                    rows.append(f"{row[0]}/{row[1]}")
        return rows

    def _get_rows_txt(self) -> List[str]:
        rows = []
        with open(self._files.INPUT) as file:
            for line in file:
                line = line.strip()
                if len(line) == 0:
                    pass
                elif len(line[0]) > 0 and line[0][0] == "#":
                    rows.append(line)
                else:
                    rows.append(line)
        return rows

    def _append_section(self, path: str):
        section_metadata = self.metadata.new_section()
        section_metadata.add_relative_file(path)
        path = f"{self._files.FILES}/{path}"
        section_metadata.FILE = path

        if not os.path.exists(path):
            raise BookfileMistakeException(f"{path} does not exist")

        file_stats = os.stat(path)
        section_metadata.CHAR_COUNT = file_stats.st_size

        with open(path, "r") as contents:
            lines = contents.readlines()
            section = RegularSection(
                lines,
                markup=self._markup,  # type: ignore [arg-type]
                fonts=self._fonts,  # type: ignore [arg-type]
                document=self._document,
                metadata=section_metadata,
            )
            section.compile()

    def _check_for_inserts(self, line: str) -> None:
        token = "INSERT:"
        insert = line.find(token)
        if insert > 0:
            path = line[insert + len(token) :].strip()
            composer = Composer(self._document)
            doc2 = Document(path)
            composer.append(doc2)

    def _check_for_metadata(self, line: str) -> None:
        token = "METADATA"
        insert = line.find(token)
        if insert > 0:
            self.append_metadata()

    def append_metadata(self):
        section = MetadataSection(
            None,
            markup=self._markup,
            fonts=self._fonts,
            document=self._document,
            metadata=self.metadata,
        )
        section.compile()



from catbook.markup import Markup
from catbook.misc import Misc
from catbook.fonts import Fonts
from catbook.files import Files
from catbook.metadata import Metadata
from catbook.section_metadata import SectionMetadata
from catbook.book_metadata import BookMetadata
from catbook.tokenizer import Tokenizer
from catbook.section import Section
from catbook.metadata_section import MetadataSection
from catbook.regular_section import RegularSection
from catbook.book import Book
from catbook.builder import Builder



from . import Markup
from . import Fonts
from . import BookMetadata
from . import Metadata
from . import Section
from docx import Document
from docx.text.paragraph import Paragraph
from docx.text.run import Run
from docx.shared import Pt
from docx.shared import RGBColor
from docx.enum.text import WD_ALIGN_PARAGRAPH, WD_BREAK
import traceback
from typing import List, Optional
from . import Tokenizer
from datetime import datetime
from typing import cast


class MetadataSection(Section):
    def __init__(
        self,
        lines: List[str],
        markup: Markup,
        fonts: Fonts,
        document: Document,
        metadata: Metadata,
    ) -> None:
        super().__init__(lines, markup, fonts, document, metadata)

    @property
    def metadata(self) -> BookMetadata:
        return cast(BookMetadata, self._metadata)

    # ============= PUBLIC STUFF HERE

    def compile(self) -> bool:
        """returns True if the lines were all added to the document"""
        p = self.doc.add_paragraph()
        run = self._add_run(p, "")
        run.font.name = self._fonts.BODY
        run.add_break(WD_BREAK.PAGE)

        run = self._add_run(p, "Book metadata", 12, True)

        table = self.doc.add_table(rows=1, cols=2)
        table.style.name = "Table Grid"

        row = table.rows[0].cells
        p = row[0].paragraphs[0]
        self._add_run(p, "Title", 10, False)
        p = row[1].paragraphs[0]
        self._add_run(p, f"{self.metadata.TITLE}", 10, False)

        row = table.add_row().cells
        p = row[0].paragraphs[0]
        self._add_run(p, "Author", 10, False)
        p = row[1].paragraphs[0]
        self._add_run(p, f"{self.metadata.AUTHOR}", 10, False)

        row = table.add_row().cells
        p = row[0].paragraphs[0]
        self._add_run(p, "Compiled on", 10, False)
        p = row[1].paragraphs[0]
        self._add_run(p, f"{datetime.now()}", 10, False)

        row = table.add_row().cells
        p = row[0].paragraphs[0]
        self._add_run(p, "Compiled from", 10, False)
        p = row[1].paragraphs[0]
        self._add_run(p, f"{self.metadata.FILES.INPUT}", 10, False)

        row = table.add_row().cells
        p = row[0].paragraphs[0]
        self._add_run(p, "Total sections", 10, False)
        p = row[1].paragraphs[0]
        self._add_run(p, f"{self.metadata.count}", 10, False)

        row = table.add_row().cells
        p = row[0].paragraphs[0]
        self._add_run(p, "Stand alone sections", 10, False)
        p = row[1].paragraphs[0]
        self._add_run(p, f"{self.metadata.stand_alone_count}", 10, False)

        row = table.add_row().cells
        p = row[0].paragraphs[0]
        self._add_run(p, "Words", 10, False)
        p = row[1].paragraphs[0]
        self._add_run(p, f"{self.metadata.word_count}", 10, False)

        row = table.add_row().cells
        p = row[0].paragraphs[0]
        self._add_run(p, "Unique words", 10, False)
        p = row[1].paragraphs[0]
        self._add_run(p, f"{self.metadata.unique_words_count()}", 10, False)

        p = self.doc.add_paragraph()
        run = self._add_run(p, "")
        p = self.doc.add_paragraph()
        run = self._add_run(p, "Compiled by catbook", 8, False, True)

        return True

    # ============= INTERNAL STUFF STARTS HERE

    def _add_run(
        self,
        p: Paragraph,
        text: str,
        size: int = 10,
        bold: bool = False,
        italic: bool = False,
    ) -> Run:
        run = p.add_run(text)
        run.bold = bold
        run.italic = italic
        run.font.name = self._fonts.BLOCK
        run.font.size = Pt(size)
        return run



from configparser import RawConfigParser
from dataclasses import dataclass
from os import path


@dataclass
class Fonts:
    QUOTE: str = "Times New Roman"
    BLOCK: str = "Courier New"
    TITLE: str = "Times New Roman"
    BODY: str = "Times New Roman"
    CONFIG: str = "config/fonts.ini"

    def __post_init__(self):
        self._config = RawConfigParser()
        self._load_config()

    def reload(self):
        self._load_config()

    def _load_config(self):
        if path.isfile(self.CONFIG):
            self._config.read(self.CONFIG)
            section = "fonts"
            try:
                self.QUOTE = self._config[section]["quote"]
            except KeyError:
                pass
            try:
                self.BLOCK = self._config[section]["block"]
            except KeyError:
                pass
            try:
                self.TITLE = self._config[section]["title"]
            except KeyError:
                pass
            try:
                self.BODY = self._config[section]["body"]
            except KeyError:
                pass
        else:
            print(f"No fonts {self.CONFIG} file found. Using default fonts.")



from . import Markup
from . import Fonts
from . import Files
from . import Misc
from . import Book
from docx import Document
from os import remove
from os.path import exists
from pathlib import Path
from typing import Optional


class BadConfigException(Exception):
    pass


class Builder:
    def __init__(self) -> None:
        self._markup: Optional[Markup] = None
        self._fonts: Optional[Fonts] = None
        self._files: Optional[Files] = None
        self._misc: Optional[Misc] = None
        self._document: Optional[Document] = None

    # ========== PUBLIC STUFF GOES HERE

    def init(self) -> None:
        if self._markup is None:
            self._markup = Markup()
        if self._fonts is None:
            self._fonts = Fonts()
        if self._files is None:
            self._files = Files()
        if self._misc is None:
            self._misc = Misc()
        self._new_document()
        self._book: Optional[Book] = None

    @property
    def book(self) -> Optional[Book]:
        if self._book is None:
            self.init()
            self.book = Book(
                files=self._files,  # type: ignore [arg-type]
                markup=self._markup,
                fonts=self._fonts,
                document=self.doc,
            )
        return self._book

    @book.setter
    def book(self, b: Book) -> None:
        self._book = b

    @property
    def doc(self) -> Optional[Document]:
        return self._document

    @doc.setter
    def doc(self, d: Document) -> None:
        self._document = d

    @property
    def files(self) -> Optional[Files]:
        return self._files

    @files.setter
    def files(self, f: Files) -> None:
        self._files = f

    @property
    def misc(self) -> Optional[Misc]:
        return self._misc

    @misc.setter
    def misc(self, m: Misc) -> None:
        self._misc = m

    def build(self):
        if None in [self._markup, self._fonts, self._files, self.doc]:
            self.init()
        if not self._validate():
            raise BadConfigException(
                f"Cannot start build with files config: {self.files}"
            )
        self._clean_output()
        self.book.create()
        self._save()
        self._reset()

    # ========== INTERNAL STUFF STARTS HERE

    def _validate(self) -> bool:
        print("Validating configuration")
        if self._files is None:
            print("No files config available")
            return False

        if self._files.INPUT is None:  # type: ignore[union-attr]
            print(f"No input file configured in {self._files.INPUT}")
            return False

        if not exists(self._files.INPUT):  # type: ignore[union-attr]
            print(f"No input file configured in {self._files}")  # type: ignore[union-attr]
            return False

        if self._files.FILES is None:  # type: ignore[union-attr]
            print(f"No files directory configured in {self._files}")
            return False

        if not exists(self._files.FILES):  # type: ignore[union-attr]
            print(f"Files directory does not exist at {self._files.FILES}")  # type: ignore[union-attr]
            return False

        return True

    def _new_document(self) -> None:
        document = Document()
        self.doc = document

    def _clean_output(self) -> None:
        """Cleans the output file location"""
        try:
            print(f"Cleaning {self._files.OUTPUT}")  # type: ignore[union-attr]
            remove(self._files.OUTPUT)  # type: ignore[union-attr]
        except FileNotFoundError:
            print(f"Nothing to clean configured in {self._files}")  # type: ignore[union-attr]

    def _reset(self) -> None:
        """removes the config and doc in case this instance gets used again"""
        self._markup = None
        self._fonts = None
        self._files = None
        self._document = None

    def _save(self):
        """not Windows safe!"""
        print(f"Saving document to {self._files.OUTPUT}")
        if "/" in self._files.OUTPUT:
            dirs = self._files.OUTPUT[0 : self._files.OUTPUT.rindex("/")]
            Path(dirs).mkdir(parents=True, exist_ok=True)
        self.doc.save(self._files.OUTPUT)



from configparser import RawConfigParser
from dataclasses import dataclass
from os import path
from typing import Optional


@dataclass
class Markup:
    """first line, first chars. a jump creates a break
    within a chapter by adding an untitled section
    separated from the preceding section by an
    indicator. often people use three widely spaced
    stars or dots as the indicator."""

    JUMP: str = "***"

    """ first line, first char. a chapter title is a grouping
        that is below book and above section """
    CHAPTER_TITLE: str = "~"

    """ first line, first char. a book title is the top
        unit. it contains chapters and sections."""
    BOOK_TITLE: str = "~~"

    """ first line, first char. forces the section to start
        on a new page """
    NEW_SECTION: str = ">"

    """ any line, first char. text that is set off from
        the rest of the quoted and unquoted paragraphs.
        could be used for inserting the text of a document
        into a narrative. """
    BLOCK: str = "|"

    """ any line, first char. a paragraph of quoted text. this
        markup is also useful for forcing a blank line. to do
        that, make it the first character of an otherwise
        blank line.  """
    QUOTED_LINE: str = '"'

    """ around any word or words. highlights the words
        between two of the marks in a paragraph. for e.g., in
        italics. Assuming | is used for both highlights and
        blocks, if a highlight begins with the first word of
        a paragraph it will look like a block. in that case
        use a double highlight mark, as in: ||some words|. """
    WORD_HIGHLIGHT: str = "|"

    CONFIG: str = "markup.ini"

    def __post_init__(self):
        self._config = RawConfigParser()
        self._load_config()

    def reload(self):
        self._load_config()

    def _load_config(self):
        if path.isfile(self.CONFIG):
            self._config.read(self.CONFIG)
            section = "markup"
            try:
                self.JUMP = self._config[section]["jump"]
            except KeyError:
                pass
            try:
                self.CHAPTER_TITLE = self._config[section]["chapter_title"]
            except KeyError:
                pass
            try:
                self.BOOK_TITLE = self._config[section]["book_title"]
            except KeyError:
                pass
            try:
                self.NEW_SECTION = self._config[section]["new_section"]
            except KeyError:
                pass
            try:
                self.BLOCK = self._config[section]["block"]
            except KeyError:
                pass
            try:
                self.QUOTED_LINE = self._config[section]["quoted_line"]
            except KeyError:
                pass
            try:
                self.WORD_HIGHLIGHT = self._config[section]["word_highlight"]
            except KeyError:
                pass
        else:
            print(f"No markup {self.CONFIG} file found. Using default markup markers.")

    def _is_quote(self, line: str, line_number: int, lines: int) -> bool:
        ll = len(line)
        if ll == 0:
            return False
        if line[0] != self.QUOTED_LINE:
            return False
        if line_number >= lines:  # why would this happen?
            return False
        return True

    def _is_block(self, line: str, line_number: int, lines: int) -> Optional[bool]:
        ll = len(line)
        if ll == 0:
            return False
        if line[0] == self.BLOCK and ll > 1 and line[1] == self.BLOCK:
            return None
        if line[0] != self.BLOCK:
            return False
        if line_number >= lines:  # why would this happen?
            return False
        return True



from . import Markup
from . import Fonts
from . import SectionMetadata
from . import Metadata
from . import Section
from docx import Document
from docx.text.paragraph import Paragraph
from docx.text.run import Run
from docx.shared import Pt
from docx.shared import RGBColor
from docx.enum.text import WD_ALIGN_PARAGRAPH, WD_BREAK
import traceback
from typing import List, Optional, cast
from . import Tokenizer


class RegularSection(Section):
    def __init__(
        self,
        lines: List[str],
        markup: Markup,
        fonts: Fonts,
        document: Document,
        metadata: Metadata,
    ) -> None:
        super().__init__(lines, markup, fonts, document, metadata)
        # self._quote: Optional[List[str]] = None

    # ============= PUBLIC STUFF HERE

    def compile(self) -> bool:
        """returns True if the lines were all added to the document"""
        reqs = [self._markup, self._fonts, self._document, self._metadata]
        if None in reqs:
            raise Exception(f"Requirements not setup: {reqs}")
        line_number = 0
        for line in self._lines:
            try:
                self._append_line(self._lines, line, line_number)
                line_number = line_number + 1
                # self._lines_count = self._lines_count + 1
            except Exception as e:
                print(f"Error in section at {line_number}: {line}: {e}")
                return False
        self.metadata.FILE_LINE_COUNT = line_number
        return True

    @property
    def doc(self) -> Document:
        return self._document

    @property
    def metadata(self) -> SectionMetadata:
        return cast(SectionMetadata, self._metadata)

    # ============= INTERNAL STUFF STARTS HERE

    def _append_quote(self) -> None:
        """adds the quote into the document"""
        if self._quote is None:  # type: ignore [has-type]
            raise Exception("No quote array found at _append_quote")
        i = len(self._quote)
        for aline in self._quote:
            if aline is None:  # make mypy happy
                continue
            p = self.doc.add_paragraph()
            paragraph_format = p.paragraph_format
            run = self._add_run(p, f"   {aline[1:]}")
            run.font.name = self._fonts.QUOTE
            run.italic = True
            i = i - 1
            if i > 0:
                paragraph_format.space_after = Pt(1)
        self._quote = None

    def _append_block(self) -> None:
        """adds the block text to the document"""
        if not self._block:
            raise Exception("No block array found at _append_block")
        i = len(self._block)
        for aline in self._block:
            if aline is None:  # make mypy happy
                continue
            p = self.doc.add_paragraph()
            paragraph_format = p.paragraph_format
            thisline = aline[1:]
            run = self._add_run(p, f"{thisline}")
            run.italic = True
            run.font.name = self._fonts.BLOCK
            run.font.size = Pt(10)
            i = i - 1
            if i > 0:
                paragraph_format.space_after = Pt(1)
        self._block = None

    def _append_output(self, lines: List[str], line: str, line_number: int) -> None:
        """writes blocks, quotes, and page breaks to the document"""
        if self._block is not None:
            # write the block
            self._append_block()
        if self._quote is not None:
            # write the quote
            self._append_quote()
        if self._part_break and self._last_line(lines, line_number):
            p = self.doc.add_paragraph()
            run = self._add_run(p, "")
            run.font.name = self._fonts.BODY
            run.add_break(WD_BREAK.PAGE)
            self._part_break = False
            self._book_break = False
            self._last_was_break = True

    def _last_line(self, lines: List[str], line_number: int) -> bool:
        """returns True if line_number indicates the last non-blank line"""
        n = len(lines)
        if line_number == n:
            return True
        for r in range(line_number + 1, n):
            if lines[r].strip() != "":
                return False
        return True

    def _append_title(self, line: str) -> None:
        """adds titles to the document"""
        self._part_break = len(line) >= 1 and line[0:1] == self._markup.CHAPTER_TITLE
        self._book_break = len(line) >= 2 and line[0:2] == self._markup.BOOK_TITLE
        simple_separator = len(line) >= 3 and line[0:3] == self._markup.JUMP
        close_part = len(line) >= 1 and line[0:1] == self._markup.NEW_SECTION
        #
        # if we're a _part_break or book_break we need a page break
        #   except if the last file was also a break. in that case
        #   we don't need another blank page
        #
        if (self._part_break and not self._last_was_break) or close_part:
            line = line[2 if self._book_break else 1 :]
            p = self.doc.add_paragraph()
            run = self._add_run(p, "")
            run.font.name = self._fonts.TITLE
            run.add_break(WD_BREAK.PAGE)
        elif self._part_break:
            line = line[2 if self._book_break else 1 :]
        p = None
        if simple_separator:
            #
            # if we're not a _part_break we might be a simple jump separator
            #
            p = self.doc.add_paragraph()
            paragraph_format = p.paragraph_format
            paragraph_format.space_before = Pt(24)
            paragraph_format.space_after = Pt(25)
            run = self._add_jump(p)
        else:
            #
            # we're not a simple separator, therefore we must be a heading of some kind
            #
            if self._part_break:
                p = self.doc.add_heading("", 1 if self._book_break else 2)
            else:
                p = self.doc.add_heading("", 3)
                paragraph_format = p.paragraph_format
                paragraph_format.space_before = Pt(30)
                paragraph_format.space_after = Pt(10)
        if self._part_break or simple_separator:
            p.alignment = WD_ALIGN_PARAGRAPH.CENTER
        if not simple_separator:
            run = self._add_run(p, line)
            run.font.name = self._fonts.BODY
            run.font.color.rgb = RGBColor.from_string("000000")

    def _handle_block(self, line, line_number, lines: int) -> Optional[bool]:
        block = self._markup._is_block(line, line_number, lines)
        if block:
            if self._block is None:
                self._block = []
            self._block.append(line)
            # pack the line into self.block
            return True
        elif block is None:
            return None
        elif self._block is not None:
            # write the block
            self._append_block()
            return False
        else:
            return False

    def _handle_quote(self, line: str, line_number: int, lines: int) -> Optional[bool]:
        quote = self._markup._is_quote(line, line_number, lines)
        if quote:
            if self._quote is None:
                self._quote = []
            self._quote.append(line)
            # we pack the line into self.quote
            return True
        elif self._quote is not None:
            # write the quote
            self._append_quote()
            return False
        else:
            return False

    def _handle_highlights(self, line):
        """adds a line that includes a highlight to the document"""
        block = 0
        p = self.doc.add_paragraph()
        while self._markup.WORD_HIGHLIGHT in line:
            start = line.index(self._markup.WORD_HIGHLIGHT)
            end = line.index(self._markup.WORD_HIGHLIGHT, start + 1)
            front = line[0:start]
            mid = line[start + 1 : end]
            back = line[end + 1 :]
            line = back
            run = self._add_run(p, "   " if block == 0 else "")
            run.font.name = self._fonts.BODY
            run = self._add_run(p, front)
            run.font.name = self._fonts.BODY
            run = self._add_run(p, mid)
            run.italic = True
            run.font.name = self._fonts.BODY
            if self._markup.WORD_HIGHLIGHT not in line:
                run = self._add_run(p, back)
                run.font.name = self._fonts.BODY
            block = block + 1

    def _add_run(self, p: Paragraph, text: str) -> Run:
        # print(f"add_run: {text} -- {self._metadata}")
        if self.metadata is not None:
            ws = self._get_words(text)
            # print(f"ws: {ws}")
            self.metadata.WORDS = ws + self.metadata.WORDS
            n = len(ws)
            self.metadata.WORD_COUNT = self.metadata.WORD_COUNT + n
        run = p.add_run(text)
        return run

    def _get_words(self, text: str) -> List[str]:
        words = Tokenizer.get_words(text)
        return words

    def _count_words(self, text: str) -> int:
        return len(self._get_words(text))

    def _add_jump(self, p: Paragraph) -> Run:
        #
        # TODO: get this text from config
        #
        run = p.add_run("*                   *                   *")
        return run

    def _append_line(self, lines: List[str], line: str, line_number: int):
        try:
            line = line.strip()
            #
            # blank lines are paragraph breaks. we take this time
            # to write out blocks, quotes, etc.
            #
            if line == "":
                self._append_output(lines, line, line_number)
                return
            #
            # titles
            #
            if line_number == 0:
                self.metadata.set_first_line(self._markup, line)
                return self._append_title(line)
            #
            # blocks
            #
            block = self._handle_block(line, line_number, len(lines))
            if block is None:
                # we found an escaped pipe: || meaning an
                # italicized word started a line
                line = line[1:]
            elif block:
                return
            #
            # quotes
            #
            quote = self._handle_quote(line, line_number, len(lines))
            if quote:
                return
            #
            # regular line
            #
            if self._markup.WORD_HIGHLIGHT in line:
                # if italics
                self._handle_highlights(line)
            else:
                p = self.doc.add_paragraph()
                run = self._add_run(p, f"   {line}")
                run.font.name = self._fonts.BODY
            self._last_was_break = False
        except Exception as e:
            print(f"Section failed at _append_line: error: {e}")
            traceback.print_exc()



from configparser import RawConfigParser
from dataclasses import dataclass
from os import path
from typing import Optional, Self


@dataclass
class Misc:
    CONFIG: str = "misc.ini"

    def __post_init__(self):
        self._config = RawConfigParser()
        self._load_config()
        self._section = "misc"
        self._index = 0
        self._name = None

    def __iter__(self) -> Self:
        return self

    def __next__(self):
        if self._name is None:
            raise StopIteration
        val = self.get(f"{self._name}_{self._index}")
        if val is None:
            self._index = 0
            self._name = None
            raise StopIteration
        else:
            self._index = self._index + 1
            return val

    def get_numbered(self, name: str) -> Self:
        """use as: values = [ _ for _ in misc.get_numbered("email")]"""
        self._name = name
        self._index = 0
        return self

    def get(self, name: str) -> Optional[str]:
        return self._config.get(self._section, name, fallback=None)

    def reload(self) -> None:
        self._load_config()

    def _load_config(self) -> None:
        if path.isfile(self.CONFIG):
            self._config.read(self.CONFIG)
        else:
            print(f"No misc {self.CONFIG} file found.")



import abc
from . import Markup
from . import Fonts
from . import SectionMetadata
from . import BookMetadata
from . import Metadata
from docx import Document
from typing import List, Optional, Union, Any


class Section(metaclass=abc.ABCMeta):
    def __init__(
        self,
        lines: List[str],
        markup: Markup,
        fonts: Fonts,
        document: Document,
        metadata: Metadata,
    ) -> None:
        self._lines: List[str] = lines
        self._fonts: Fonts = fonts
        self._markup: Markup = markup
        self._document: Document = document
        self._block: Optional[List[Optional[str]]] = None
        self._quote: Optional[List[Optional[str]]] = None
        # self._lines_count: int = 0
        self._part_break: bool = False
        self._last_was_break = False
        self._metadata = metadata

    @property
    def doc(self) -> Document:
        return self._document

    @abc.abstractmethod
    def compile(self) -> bool:
        pass
